@@ -16,6 +16,7 @@
 
 #include <sys/types.h>
 #include <sys/stat.h>
+#include <sys/queue.h>
 
 #include <limits.h>
 #include <stdio.h>
@@ -212,8 +213,6 @@ packed_first_parent_traversal(int *ncommits_traversed,
 
 	/* Add all traversed commits to the graph... */
 	STAILQ_FOREACH(qid, &traversed_commits, entry) {
-		struct got_commit_graph_node *node;
-
 		if (got_object_idset_contains(graph->open_branches, &qid->id))
 			continue;
 		if (got_object_idset_contains(graph->node_ids, &qid->id))
@@ -228,7 +227,7 @@ packed_first_parent_traversal(int *ncommits_traversed,
 			break;
 		}
 
-		err = add_node(&node, graph, &qid->id, repo);
+		err = got_object_idset_add(graph->node_ids, &qid->id, NULL);
 		if (err)
 			break;
 	}
@@ -254,6 +253,7 @@ advance_branch(struct got_commit_graph *graph, struct got_object_id *commit_id,
 {
 	const struct got_error *err;
 	struct got_object_qid *qid;
+	struct got_object_id *merged_id = NULL;
 
 	err = close_branch(graph, commit_id);
 	if (err)
@@ -289,13 +289,16 @@ advance_branch(struct got_commit_graph *graph, struct got_object_id *commit_id,
 	 * which do not contribute any content to this path.
 	 */
 	if (commit->nparents > 1 && !got_path_is_root_dir(graph->path)) {
-		struct got_object_id *merged_id, *prev_id = NULL;
+		err = got_object_id_by_path(&merged_id, repo, commit, graph->path);
+		if (err && err->code != GOT_ERR_NO_TREE_ENTRY)
+			return err;
+		/* The requested path does not exist in this merge commit. */
+	}
+	if (commit->nparents > 1 && !got_path_is_root_dir(graph->path) &&
+	    merged_id != NULL) {
+		struct got_object_id *prev_id = NULL;
 		int branches_differ = 0;
 
-		err = got_object_id_by_path(&merged_id, repo, commit,
-		    graph->path);
-		if (err)
-			return err;
 
 		STAILQ_FOREACH(qid, &commit->parent_ids, entry) {
 			struct got_object_id *id = NULL;
@@ -445,10 +448,12 @@ add_branch_tip(struct got_object_id *commit_id, void *data, void *arg)
 		return err;
 
 	err = add_node(&new_node, a->graph, commit_id, a->repo);
-	if (err)
+	if (err) {
+		got_object_commit_close(commit);
 		return err;
+	}
 
-	a->tips[a->ntips].commit_id = new_node ? &new_node->id : NULL;
+	a->tips[a->ntips].commit_id = &new_node->id;
 	a->tips[a->ntips].commit = commit;
 	a->tips[a->ntips].new_node = new_node;
 	a->ntips++;
@@ -517,22 +522,33 @@ fetch_commits_from_open_branches(struct got_commit_graph *graph,
 				break;
 			continue;
 		}
-		if (changed)
+		if (changed) {
 			add_node_to_iter_list(graph, new_node,
 			    got_object_commit_get_committer_time(commit));
+			arg.tips[i].new_node = NULL;
+		}
 		err = advance_branch(graph, commit_id, commit, repo);
 		if (err)
 			break;
 	}
 done:
-	for (i = 0; i < arg.ntips; i++)
+	for (i = 0; i < arg.ntips; i++) {
 		got_object_commit_close(arg.tips[i].commit);
+		free(arg.tips[i].new_node);
+	}
 	return err;
 }
 
 void
 got_commit_graph_close(struct got_commit_graph *graph)
 {
+	struct got_commit_graph_node *node;
+
+	while ((node = TAILQ_FIRST(&graph->iter_list))) {
+		TAILQ_REMOVE(&graph->iter_list, node, entry);
+		free(node);
+	}
+
 	if (graph->open_branches)
 		got_object_idset_free(graph->open_branches);
 	if (graph->node_ids)
@@ -579,15 +595,13 @@ got_commit_graph_iter_start(struct got_commit_graph *graph,
 }
 
 const struct got_error *
-got_commit_graph_iter_next(struct got_object_id **id,
+got_commit_graph_iter_next(struct got_object_id *id,
     struct got_commit_graph *graph, struct got_repository *repo,
     got_cancel_cb cancel_cb, void *cancel_arg)
 {
 	const struct got_error *err = NULL;
 	struct got_commit_graph_node *node;
 
-	*id = NULL;
-
 	node = TAILQ_FIRST(&graph->iter_list);
 	if (node == NULL) {
 		/* We are done iterating, or iteration was not started. */
@@ -602,11 +616,43 @@ got_commit_graph_iter_next(struct got_object_id **id,
 			return err;
 	}
 
-	*id = &node->id;
+	memcpy(id, &node->id, sizeof(*id));
+
 	TAILQ_REMOVE(&graph->iter_list, node, entry);
+	free(node);
 	return NULL;
 }
 
+static const struct got_error *
+find_yca_add_id(struct got_object_id **yca_id, struct got_commit_graph *graph,
+    struct got_object_idset *commit_ids, struct got_repository *repo,
+    got_cancel_cb cancel_cb, void *cancel_arg)
+{
+	const struct got_error *err = NULL;
+	struct got_object_id id;
+
+	err = got_commit_graph_iter_next(&id, graph, repo, cancel_cb,
+	    cancel_arg);
+	if (err)
+		return err;
+
+	if (got_object_idset_contains(commit_ids, &id)) {
+		*yca_id = got_object_id_dup(&id);
+		if (*yca_id == NULL)
+			err = got_error_from_errno("got_object_id_dup");
+		return err;
+	}
+
+	return got_object_idset_add(commit_ids, &id, NULL);
+}
+
+/*
+ * Sets *yca_id to the youngest common ancestor of commit_id and
+ * commit_id2. Returns got_error(GOT_ERR_ANCESTRY) if they have no
+ * common ancestors.
+ *
+ * If first_parent_traversal is nonzero, only linear history is considered.
+ */
 const struct got_error *
 got_commit_graph_find_youngest_common_ancestor(struct got_object_id **yca_id,
     struct got_object_id *commit_id, struct got_object_id *commit_id2,
@@ -643,8 +689,6 @@ got_commit_graph_find_youngest_common_ancestor(struct got_object_id **yca_id,
 		goto done;
 
 	for (;;) {
-		struct got_object_id *id = NULL, *id2 = NULL;
-
 		if (cancel_cb) {
 			err = (*cancel_cb)(cancel_arg);
 			if (err)
@@ -652,7 +696,7 @@ got_commit_graph_find_youngest_common_ancestor(struct got_object_id **yca_id,
 		}
 
 		if (!completed) {
-			err = got_commit_graph_iter_next(&id, graph, repo,
+			err = find_yca_add_id(yca_id, graph, commit_ids, repo,
 			    cancel_cb, cancel_arg);
 			if (err) {
 				if (err->code != GOT_ERR_ITER_COMPLETED)
@@ -660,10 +704,12 @@ got_commit_graph_find_youngest_common_ancestor(struct got_object_id **yca_id,
 				err = NULL;
 				completed = 1;
 			}
+			if (*yca_id)
+				break;
 		}
 
 		if (!completed2) {
-			err = got_commit_graph_iter_next(&id2, graph2, repo,
+			err = find_yca_add_id(yca_id, graph2, commit_ids, repo,
 			    cancel_cb, cancel_arg);
 			if (err) {
 				if (err->code != GOT_ERR_ITER_COMPLETED)
@@ -671,32 +717,7 @@ got_commit_graph_find_youngest_common_ancestor(struct got_object_id **yca_id,
 				err = NULL;
 				completed2 = 1;
 			}
-		}
-
-		if (id) {
-			if (got_object_idset_contains(commit_ids, id)) {
-				*yca_id = got_object_id_dup(id);
-				if (*yca_id)
-					break;
-				err = got_error_from_errno("got_object_id_dup");
-				break;
-
-			}
-			err = got_object_idset_add(commit_ids, id, NULL);
-			if (err)
-				break;
-		}
-		if (id2) {
-			if (got_object_idset_contains(commit_ids, id2)) {
-				*yca_id = got_object_id_dup(id2);
-				if (*yca_id)
-					break;
-				err = got_error_from_errno("got_object_id_dup");
-				break;
-
-			}
-			err = got_object_idset_add(commit_ids, id2, NULL);
-			if (err)
+			if (*yca_id)
 				break;
 		}
 
@@ -704,7 +725,6 @@ got_commit_graph_find_youngest_common_ancestor(struct got_object_id **yca_id,
 			err = got_error(GOT_ERR_ANCESTRY);
 			break;
 		}
-
 	}
 done:
 	got_object_idset_free(commit_ids);
