@@ -141,12 +141,16 @@ got_privsep_recv_imsg(struct imsg *imsg, struct imsgbuf *ibuf,
 			return got_error_from_errno("imsg_get");
 	}
 
-	if (imsg->hdr.len < IMSG_HEADER_SIZE + min_datalen)
-		return got_error(GOT_ERR_PRIVSEP_LEN);
-
 	if (imsg->hdr.type == GOT_IMSG_ERROR) {
 		size_t datalen = imsg->hdr.len - IMSG_HEADER_SIZE;
-		return recv_imsg_error(imsg, datalen);
+		err = recv_imsg_error(imsg, datalen);
+		imsg_free(imsg);
+		return err;
+	}
+
+	if (imsg->hdr.len < IMSG_HEADER_SIZE + min_datalen) {
+		imsg_free(imsg);
+		return got_error(GOT_ERR_PRIVSEP_LEN);
 	}
 
 	return NULL;
@@ -1183,8 +1187,7 @@ got_privsep_recv_obj(struct got_object **obj, struct imsgbuf *ibuf)
 {
 	const struct got_error *err = NULL;
 	struct imsg imsg;
-	const size_t min_datalen =
-	    MIN(sizeof(struct got_imsg_error), sizeof(struct got_imsg_object));
+	const size_t min_datalen = sizeof(struct got_imsg_object);
 
 	*obj = NULL;
 
@@ -1405,9 +1408,7 @@ got_privsep_recv_commit(struct got_commit_object **commit, struct imsgbuf *ibuf)
 	const struct got_error *err = NULL;
 	struct imsg imsg;
 	size_t datalen;
-	const size_t min_datalen =
-	    MIN(sizeof(struct got_imsg_error),
-	    sizeof(struct got_imsg_commit_object));
+	const size_t min_datalen = sizeof(struct got_imsg_commit_object);
 
 	*commit = NULL;
 
@@ -1592,9 +1593,7 @@ const struct got_error *
 got_privsep_recv_tree(struct got_tree_object **tree, struct imsgbuf *ibuf)
 {
 	const struct got_error *err = NULL;
-	const size_t min_datalen =
-	    MIN(sizeof(struct got_imsg_error),
-	    sizeof(struct got_imsg_tree_object));
+	const size_t min_datalen = sizeof(struct got_imsg_tree_object);
 	struct got_imsg_tree_object *itree;
 	int nentries = 0;
 
@@ -1849,9 +1848,7 @@ got_privsep_recv_tag(struct got_tag_object **tag, struct imsgbuf *ibuf)
 	struct imsg imsg;
 	struct got_imsg_tag_object *itag;
 	size_t len, datalen;
-	const size_t min_datalen =
-	    MIN(sizeof(struct got_imsg_error),
-	    sizeof(struct got_imsg_tag_object));
+	const size_t min_datalen = sizeof(struct got_imsg_tag_object);
 
 	*tag = NULL;
 
@@ -2206,8 +2203,7 @@ got_privsep_recv_gitconfig_int(int *val, struct imsgbuf *ibuf)
 	const struct got_error *err = NULL;
 	struct imsg imsg;
 	size_t datalen;
-	const size_t min_datalen =
-	    MIN(sizeof(struct got_imsg_error), sizeof(int));
+	const size_t min_datalen = sizeof(int);
 
 	*val = 0;
 
@@ -2484,8 +2480,7 @@ got_privsep_recv_gotconfig_remotes(struct got_remote_repo **remotes,
 	size_t datalen;
 	struct got_imsg_remotes iremotes;
 	struct got_imsg_remote iremote;
-	const size_t min_datalen =
-	    MIN(sizeof(struct got_imsg_error), sizeof(iremotes));
+	const size_t min_datalen = sizeof(iremotes);
 
 	*remotes = NULL;
 	*nremotes = 0;
@@ -2525,8 +2520,7 @@ got_privsep_recv_gotconfig_remotes(struct got_remote_repo **remotes,
 
 	while (*nremotes < iremotes.nremotes) {
 		struct got_remote_repo *remote;
-		const size_t min_datalen =
-		    MIN(sizeof(struct got_imsg_error), sizeof(iremote));
+		const size_t min_datalen = sizeof(iremote);
 		int i;
 
 		err = got_privsep_recv_imsg(&imsg, ibuf, min_datalen);
@@ -3510,10 +3504,8 @@ got_privsep_recv_painted_commits(struct got_object_id_queue *new_ids,
 
 	for (;;) {
 		err = got_privsep_recv_imsg(&imsg, ibuf, 0);
-		if (err){
-			imsg_free(&imsg);
+		if (err)
 			return err;
-		}
 
 		datalen = imsg.hdr.len - IMSG_HEADER_SIZE;
 		if (imsg.hdr.type == GOT_IMSG_COMMIT_PAINTING_DONE) {
