@@ -34,6 +34,7 @@
 #include <zlib.h>
 #include <err.h>
 
+#include "got_compat.h"
 #include "got_error.h"
 #include "got_object.h"
 #include "got_path.h"
@@ -134,7 +135,7 @@ recv_packfd(int *packfd, struct imsgbuf *ibuf)
 		goto done;
 	}
 
-	*packfd = imsg.fd;
+	*packfd = imsg_get_fd(&imsg);
 done:
 	imsg_free(&imsg);
 	return err;
@@ -211,7 +212,6 @@ send_their_ref(struct imsgbuf *ibuf, struct got_object_id *refid,
 	if (imsg_add(wbuf, refname, reflen) == -1)
 		return got_error_from_errno("imsg_add SEND_REMOTE_REF");
 
-	wbuf->fd = -1;
 	imsg_close(ibuf, wbuf);
 	return got_privsep_flush_imsg(ibuf);
 }
@@ -295,7 +295,6 @@ send_ref_status(struct imsgbuf *ibuf, const char *refname, int success,
 	if (imsg_add(wbuf, errmsg, errmsglen) == -1)
 		return got_error_from_errno("imsg_add SEND_REF_STATUS");
 
-	wbuf->fd = -1;
 	imsg_close(ibuf, wbuf);
 	return got_privsep_flush_imsg(ibuf);
 }
@@ -658,7 +657,7 @@ main(int argc, char **argv)
 		goto done;
 	}
 	memcpy(&send_req, imsg.data, sizeof(send_req));
-	sendfd = imsg.fd;
+	sendfd = imsg_get_fd(&imsg);
 	imsg_free(&imsg);
 
 	if (send_req.verbosity > 0)
